"""Exports Animation and AnimationPlayer used to display sprite-based 
animated characters in the game."""

from osm import agent

class Animation(agent.Agent):
    """Plays animation by displaying images from a list of N frames.
    The image list is a list of pairs [[time to display,image name],...].
    Animation has recognised state 'playing' (meaning has action for 
    this state) and responds to commands 'reset' and 'start'.
    Animation does not hold images themselves. It refers to them by names.
    Mapping from a name to the image is done by ActorVisual.
    """
    
    def __init__(self, name=None, images=None, visual=None):
        self.animation_debug = 0
        agent.Agent.__init__(self)
        self.visual = visual
        #print "animation ",self.name," got visual ",self.visual
        self.images = images
        self.currImage = None
        self.frameTimer = 0
        self.timesPlayed = 0
        self.actions["playing"] = self.Action_Playing
        self.messageHandlers["start"] = self.Handler_OnStart
        self.messageHandlers["reset"] = self.Handler_OnReset
        self.setImage(0)

    def trace(self, args):
        if (self.animation_debug):
            print self.name+": "+str(args)

    def Handler_OnReset(self,msg):
        """Resets current image and frameTimer."""
        self.frameTimer = 0
        self.timesPlayed = 0
        self.setImage(0)
        self.setState("playing")

    def Handler_OnStart(self,msg):
        """Sets state to playing."""
        self.trace("animation OnStart")
        self.setState("playing")

    def Action_Playing(self):
        """Updates frameTimer and switches images when frameTimer reaches specified values."""
        self.frameTimer += 1
        self.trace(("time = ",self.frameTimer))
        if (len(self.images)==0):
            self.setState("empty animation")
            print "empty animation!!!"
            return None
        if (self.frameTimer >= self.images[self.currImage][0]):
            self.nextImage()

    def setImage(self,id):
        """Requests visual (renderer) to display an image."""
        self.currImage = id
        if (len(self.images)==0):
            return
        image = self.images[self.currImage][1]
        if (self.visual!=None):
            self.trace("requesting image: "+str(id))
            self.visual.sendCommand("show image",paramsDict={"image":image})
            self.frameTimer = 0

    def nextImage(self):
        """Advances image number and wraps it around N images."""
        self.currImage += 1
        if (self.currImage>=len(self.images)):
            self.timesPlayed += 1
            self.currImage = 0
        self.setImage(self.currImage)
        self.frameTimer = 0

class AnimationPlayer(agent.Agent):
    """Keeps a list of animations and can play them by name. 
    Command 'start' specifies which animation to play and initiates
    playback. Update function for the current animation is called 
    locally, by AnimationPlayer.'"""
    
    def __init__(self, name=None, anims=None):
        self.animationmanager_debug = 0
        agent.Agent.__init__(self,name = name)
        self.anims = anims
        self.currAnim = None
        self.actions["playing"] = self.Action_Playing
        self.messageHandlers["start"] = self.Handler_OnStart

    def debug_report(self, txt):
        print "anim player report ",txt
        for anim in self.anims.keys():
            print ">>",anim,":",self.anims[anim].visual

    def trace(self, args):
        if (self.animationmanager_debug):
            print args

    def Handler_OnStart(self,msg):
        """Sets state to playing. The animation is specified in message
        parameter 'name'."""
        self.trace("animationManager OnStart")
        animName = msg.paramByName("name")
        if (self.anims.has_key(animName)):
            self.currAnim = self.anims[animName]
            self.currAnim.name = animName
            self.currAnim.sendCommand("reset")
            self.currAnim.sendCommand("start")
            self.setState("playing")
        else:
            self.currAnim = None
            self.setState("idle")
            self.trace("animation not found "+animName)

    def Action_Playing(self):
        """Keeps updating animation and clears currAnim inBox 
        after each update."""
        if (self.currAnim):
            self.currAnim.update()
            self.currAnim.clearInBox()

