"""Callable module that runs games and displays some Tkinter-based UI."""

import Image
import ImageTk
import Tkinter
import tkFileDialog
import game
import xmlscript
import os
import datapath
import Pmw
import scriptbrowser
import objectbrowser

class GameUI:
    """UI for the game object. Sends timed message to itself to update UI and run the game update.""" 
    def __init__(self, dataPath):
        self.dataPath = dataPath
        self.mapSize = (790,700)
        self.root=Tkinter.Tk()
        self.root.title("OSM Test v0.502")
        self.buildControlPanel()
        self.buildNotebook()
        self.canvas = Tkinter.Canvas(self.gamePage,width=self.mapSize[0],height=self.mapSize[1])
        self.createScriptEditor()
        self.createObjectBrowser()
        self.game = None
        self.gamefile = None
        self.gameRunning = 0 #true if gameui is expecting next timer event to update game
        self.selectedAgent = None
        self.registerUserInputCallbacks() 
        self.canvas.pack(fill=Tkinter.BOTH,expand=Tkinter.YES)
        self.root.mainloop()

    def registerUserInputCallbacks(self):
        """Binds Tkinter mouse and keyboard events to handlers defined in gameui."""
        self.canvas.bind("<Button-1>", self.canvasClicked)
        self.root.bind("<Key>", self.keyPressed)

    def updateSelectedAgentDisplay(self):
        """Updates info of the selected object."""
        if (self.selectedAgent!=None):
            self.agentLabel.configure(text=self.selectedAgent.name)
        else:
            self.agentLabel.configure(text="No selection")

    def selectObject(self,pos):
        """Requests the game to select an object at specified position."""
        if (self.game):
            self.game.sendCommand("select",paramsDict={"pos":pos})
            for msg in self.game.outBox:
                if (msg.type=="selected"):
                    self.selectedAgent = msg.dict["selection"]
                    self.updateSelectedAgentDisplay()
        
    def canvasClicked(self, event):
        """Handler for mouse clicks on canvas (the game field). When called, it sends 
        commands to the game object."""
        if (self.game==None):
            return
        if (self.game.currState=="running"):
            self.game.sendCommand("goto",paramsDict={"destination":complex(event.x,event.y)})
        else: #paused
            self.selectObject(complex(event.x,event.y))

    def keyPressed(self,event):
        """Handler for keyboard events. Redirects them to the game object."""
        #print "event.keycode=",event.keycode
        if (self.game==None):
            return
        self.game.sendCommand("keyboard",paramsDict={"event":event})
    
    def test1(self):
        """Test function."""
        pass
        
    def test2(self):
        """Test function."""
        self.game.player.sendCommand("getknocked")
    
    def buildNotebook(self):
        self.notebook = Pmw.NoteBook(self.root, tabpos = 'n', hull_width = self.mapSize[0]+10,
                hull_height = self.mapSize[1]+40,)
        self.notebook.pack(fill=Tkinter.BOTH, expand=Tkinter.YES)
        self.gamePage = self.notebook.add('Game')
        self.notebook.tab('Game')#.focus_set()
        self.objectsPage = self.notebook.add('Objects')
        self.notebook.tab('Objects')#.focus_set()
        self.scriptPage = self.notebook.add('Scripts')
        self.notebook.tab('Scripts')#.focus_set()
        self.animPage = self.notebook.add('Animations')
        self.notebook.tab('Animations')#.focus_set()
    
    def createScriptEditor(self):
        self.scriptEditor = scriptbrowser.ScriptEditor(".\\games",self.scriptPage)
    
    def createObjectBrowser(self):
        self.objectBrowser = objectbrowser.ObjectBrowser(self.objectsPage,None)
    
    def buildControlPanel(self):
        """Creates UI elements on the right side of the game window."""
        self.controlPanel = Tkinter.Frame(self.root, width = 200, height= self.mapSize[1])
        playFrame = Tkinter.Frame(self.controlPanel, width = 198, relief=Tkinter.GROOVE, borderwidth=2)
        Tkinter.Button(playFrame,text="Open",command=self.loadGame).pack(side=Tkinter.LEFT)
        Tkinter.Button(playFrame,text="Restart",command=self.restartGame).pack(side=Tkinter.LEFT)
        Tkinter.Button(playFrame,text="Run",command=self.run).pack(side=Tkinter.LEFT)
        Tkinter.Button(playFrame,text="Pause",command=self.pause).pack(side=Tkinter.LEFT)
        Tkinter.Button(playFrame,text="Step",command=self.step).pack(side=Tkinter.LEFT)
        playFrame.pack(anchor=Tkinter.NW,fill=Tkinter.BOTH)

        testFrame = Tkinter.Frame(self.controlPanel, width = 198, relief=Tkinter.GROOVE, borderwidth=2)
        Tkinter.Button(testFrame,text="Test 1",command=self.test1).pack(side=Tkinter.LEFT)
        Tkinter.Button(testFrame,text="Test 2",command=self.test2).pack(side=Tkinter.LEFT)
        testFrame.pack(anchor=Tkinter.NW,fill=Tkinter.BOTH)
        
        self.statusString = "No game loaded."
        self.statusLabel = Tkinter.Label(self.controlPanel, text=self.statusString, relief=Tkinter.SUNKEN, justify=Tkinter.LEFT)
        self.statusLabel.pack(anchor=Tkinter.NW, fill=Tkinter.BOTH)
        self.agentLabel = Tkinter.Label(self.controlPanel, text="No selection", relief=Tkinter.SUNKEN, justify=Tkinter.LEFT)
        self.agentLabel.pack(anchor=Tkinter.NW, fill=Tkinter.BOTH)
        
        self.controlPanel.pack(side = Tkinter.RIGHT, fill=Tkinter.BOTH)
        
    def update(self):
        """Calls itself every X msec."""
        self.gameRunning = 0
        self.game.update()
        self.updateSelectedAgentDisplay()
        if (self.game.currState=="running"):
            self.root.after(10,self.update)
            self.gameRunning = 1

    def run(self, nSteps=-1):
        """Sends command GameRun to the game object and creates 
        timer event to continue running in cycles."""
        if (self.game!=None) and (nSteps!=0):
            self.gameRunning = 1
            self.root.after(10, self.update)
            self.game.sendCommand("run", paramsDict={"nSteps":nSteps})
    
    def pause(self):
        """Pauses the game by setting stepsToRun to 0 and by NOT generating timer event."""
        if (self.game!=None):
            self.game.stepsToRun = 0
            self.gameRunning = 0

    def restartGame(self):
        """Reloads the game from a valid cached game path."""
        if (self.gamefile):
            del self.game
            self.game = game.Game(self.root,self.canvas)
            datapath.gameroot = os.path.splitext(self.gamefile)[0]+"\\"  
            self.game.setScript(xmlscript.scriptFromFile(self.gamefile))
            self.agent_debug = 1
            self.game.sendCommand("initialize")
            self.gameRunning = 0
            #setup script editor
            self.scriptEditor.openGame(self.gamefile)
            self.objectBrowser.setGame(self.game)
            
    def loadGame(self):
        """Asks for the game file name and calls restartGame when a 
        valid file name is provided."""
        self.gamefile = tkFileDialog.askopenfilename(filetypes=[("all files", "*.xml")],initialdir=".\\games")
        try:
            self.statusLabel.configure(text="Game loaded from file:\n"+self.gamefile)
        except:
            self.statusLabel.configure(text="Could not open File: "+self.gamefile)
            self.gamefile = None
            return 
        self.restartGame()
    
    def step(self):
        """Runs single step of game update."""
        self.run(1)
        self.gameRunning = 0
        
    def updateUICbk(self):
        """Callback for the game objects. Calling this function they can update display information stored in
        self.game.statusString. (This function is not working in the latest version yet.)"""
        status = self.game.statusString + ": ("+str(self.game.currFrame)+")"#+str(self.game.player.motor.blocking)
        self.statusLabel.configure(text=status)

if (__name__=="__main__"):
    g = GameUI(".\\")
    g.run(0)
