"""Exports class Movement that keeps track of Actor's position
and calculates a desired transition for the next frame."""

from osm import agent

class Movement(agent.Agent):
    """A movement calculator class. It keeps current position of an object.
    On update, it takes current position, destination and returns a message 
    into the outBox. The message contains desired translation so that the
    object can move to its desitnation."""
    def __init__(self,startPos=10+10j,name="Movement"):
        agent.Agent.__init__(self,name)
        self.speed = 3
        self.currPos = startPos
        self.prevPos = startPos
        self.destPos = startPos
        self.step = 0+0j 
        self.actions["going"] = self.Action_Going
        self.messageHandlers["goto"] = self.Handler_OnGoTo
        self.messageHandlers["setposition"] = self.Handler_OnSetPos
        self.messageHandlers["disable"] = self.Handler_OnDisable
        self.messageHandlers["enable"] = self.Handler_OnEnable
        self.messageHandlers["setspeed"] = self.Handler_OnSetSpeed
        self.messageHandlers["resolvedTranslation"] = self.Hanlder_OnResolvedTranslation
        self.setState("idle")

    def Handler_OnDisable(self,msg):
        """Disables movement (when actor is blocking or knocked down).
        Movement ignores all goto messages received while disabled."""
        self.setState("disabled")
        
    def Handler_OnEnable(self,msg):
        """Enables movement so it can handle incoming messages again."""
        self.setState("idle")

    def Handler_OnGoTo(self,msg):
        """Sets destination and switches to 'going' state."""
        if (self.currState=="disabled"):
            self.sendReport("disabled")
            return
        self.destPos = msg.paramByName("destination")
        self.setState("going")
        #print "movement received goto command, dest=",self.destPos
        
    def destinaionReached(self):
        """Helper function, returns 1 if actor is close enough to its destination."""
        if (int(self.currPos.imag)==int(self.destPos.imag)) and (int(self.currPos.real)==int(self.destPos.real)):
            return 1
        return 0
    
    def Action_Going(self):
        """Calculate desired displacement and post it to outBox. 
        Ideally the actual position modification should be delayed 
        and handled in a dedicated collision response system."""
        #actual modification delayed until collision response handler
        #print "going..."
        if (not self.destinaionReached()):
            #print "destination NOT reached ",self.destPos, self.currPos
            self.step = self.destPos - self.currPos
            len = abs(self.step)
            mult = self.speed
            if (len < self.speed):
                mult = 1 #can finish move in one frame
            self.step *= mult /len
            maxDelta = abs(self.step.real)
            if (maxDelta < abs(self.step.imag)):
                maxDelta = abs(self.step.imag)
            if (maxDelta<1):
                self.step /= maxDelta
            self.sendReport("desiredTranslation", delay=1, paramsDict={"delta":self.step, "currPos":self.currPos})
        else:
            #print "destination reached ",self.destPos, self.currPos
            self.sendReport("idle")
            self.setState("idle")
    
    def Handler_OnSetPos(self,msg):
        """Handler for the command forcing to set object's current position."""
        self.prevPos = self.currPos
        self.currPos = msg.paramByName("newPos")
        self.currPos = complex(int(self.currPos.real),int(self.currPos.imag))

    def Handler_OnSetSpeed(self,msg):
        self.speed = msg.paramByName("speed")

    def Hanlder_OnResolvedTranslation(self,msg):
        newDelta = msg.paramByName("correctedDelta")
        self.sendReport("translation", paramsDict={"delta":newDelta, "currPos":self.currPos})
        newPos = self.currPos + newDelta
        self.sendCommand("setposition",paramsDict={"newPos":newPos})
        